@@ -232,6 +232,7 @@ shutdown(Pid) ->
 
 init([UID]) ->
     set_logging_prefix(UID),
+    ?LOG_DEBUG("Creating E-RAB FSM (UID ~p)", [UID]),
     %% request a unieue SEID for this E-RAB FSM
     {ok, SEID} = pfcp_peer:seid_alloc(),
     {ok, erab_wait_setup_req,
@@ -546,6 +547,17 @@ erab_wait_release_rsp(cast, erab_release_ind,
     ?LOG_INFO("Rx E-RAB RELEASE Ind while releasing (ignored)"),
     keep_state_and_data;
 
+%% late E-RAB SETUP Rsp (race condition)
+%% see test_initial_context_setup_rsp_during_release
+erab_wait_release_rsp({call, From},
+                      {erab_setup_rsp, U2A},
+                      #erab_state{u2a = undefined} = S) ->
+    ?LOG_NOTICE("Rx late E-RAB SETUP Rsp (U2A ~s)", [pp(U2A)]),
+    Reply = {ok, S#erab_state.c2u},
+    {keep_state,
+     S#erab_state{u2a = U2A},
+     {reply, From, Reply}};
+
 erab_wait_release_rsp(Event, EventData, S) ->
     handle_event(?FUNCTION_NAME, Event, EventData, S).
 
@@ -598,7 +610,7 @@ terminate(Reason, State, S) ->
 %% set process metadata for the logger
 set_logging_prefix(UID) ->
     Prefix = io_lib:format("E-RAB ~p", [UID]),
-    logger:set_process_metadata(#{prefix => Prefix}).
+    osmo_s1gw:set_log_prefix(Prefix).
 
 
 -spec session_establish_req(erab_state()) -> pfcp_peer:pfcp_session_rsp().
